def hamming(a: str, b: str) -> int:
    """Compute the Hamming distance between two strings."""

def levenshtein(a: str, b: str) -> int:
    """Compute the Levenshtein distance between two strings."""

def needleman_wunsch(a: str, b: str) -> int:
    """Compute the Needleman-Wunsch distance between two strings."""
